#ifndef STREAMPP_H
#define STREAMPP_H
                    
/* WICHTIG: #define-Anweisungen muessen aus Portabilitaetsgruenden
	    immer in Spalte Null des Quelltextes beginnen.         */

/* fuer C++ Translator cfront von AT&T */
//#define cfrontCpp

/* fuer Borland C++ und Microsoft Visual C++ */
#define  BorlandCpp_MicrosoftCpp

/* fuer Zortech C++ */
//#define  ZortechCpp

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>


/******************************************************************/
/*********** Systemabhaengige Einstellungen ***********************/
/******************************************************************/

/*
 RegClass(name)
 Registriert den Klassennamen mit einem Pointer auf die zugehoerige
 Build-Methode im ClassManager (ermoeglicht Zugriff auf Build-Methode
 ueber den Klassennamen)
*/

/*
 DefNameBuild(name)
 Erstellt die public Methoden Classname() und Build(). Classname()
 liefert den Klassennamen als String (char*) zurueck. Die static-
 Methode Build(TIRefStream* load_stream) ruft den Build-Konstruktor
 dieser Klasse auf. Dieses Makro muss in der Klassendefiniton jeder
 Klasse mit persistenzfaehigen Instanzen aufgerufen werden.(Nach dem
 Makroaufruf gelten alle Elemente bis zur naechsten private- oder
 protected-Anweisung als public.)
*/


/******************************************************************/
/***** Initialisierungsblock fuer Borland C++/ Microsoft C++ ******/
/***** (DOS/Windows) **********************************************/
/******************************************************************/

#ifdef BorlandCpp_MicrosoftCpp
#define RegClass(name) ClassManager.Insert(new TAssociation(#name,(Pvoid)name::Build));
#define DefNameBuild(name) public: virtual Pchar Classname(){return #name;};  static PTPersistent Build(TIRefStream* load_stream) { return new name(load_stream);};
#define DefClass(name) class name; typedef name* P##name;

#define ReverseByteSequencesForPC
/* Beim PC sind gegenueber anderen Maschinen High- und Low-Byte
   vertauscht. Die Definition von ReverseByteSequenceForPC signalisiert
   hier, dass beim Speichern/Laden die Bytefolgen jeweils vertauscht
   werden. */

const int MaxCollectionSize = 16380;
/* 16380 weil auf dem PC die 64kByte-Grenze fuer eine Datenstruktur nicht
   ueberschritten werden kann und ein far-Pointer 4 Bytes benoetigt */

#define FileInOpenMode ios::in|ios::binary
#define FileOutOpenMode ios::out|ios::binary
#endif


/******************************************************************/
/***** Initialisierungsblock fuer Zortech C++ *********************/
/******************************************************************/

#ifdef ZortechCpp
#define RegClass(name) ClassManager.Insert(new TAssociation(#name,(Pvoid)name::Build));
#define DefNameBuild(name) public: virtual Pchar Classname(){return #name;};  static PTPersistent Build(TIRefStream* load_stream) { return new name(load_stream);};
#define DefClass(name) class name; typedef name* P##name;

#define ReverseByteSequencesForPC
/* Beim PC sind gegenueber anderen Maschinen High- und Low-Byte
   vertauscht. Die Definition von ReverseByteSequenceForPC signalisiert
   hier, dass beim Speichern/Laden die Bytefolgen jeweils vertauscht
   werden. */

const int MaxCollectionSize = 16380;
/* 16380 weil auf dem PC die 64kByte-Grenze fuer eine Datenstruktur nicht
   ueberschritten werden kann und ein far-Pointer 4 Bytes benoetigt */

#define FileInOpenMode ios::in
#define FileOutOpenMode ios::out
#endif


/******************************************************************/
/**** Initialisierungsblock fuer cfront C++ (Unix) *****************/
/******************************************************************/

#ifdef cfrontCpp
#define RegClass(name) ClassManager.Insert(new TAssociation("name",(Pvoid) name::Build));
#define DefNameBuild(name)public: virtual Pchar Classname(){return "name";}; static PTPersistent Build(TIRefStream* load_stream) { return new name(load_stream);};

#define DefClass(name) class name; typedef name* P/**/name;

const int MaxCollectionSize = 16380;
/* Aus Kompatibilitaetsgruenden die selbe Obergrenze wie beim PC. */

#define FileInOpenMode ios::in
#define FileOutOpenMode ios::out
#endif

/******************************************************************/
/*********** Schwebende Definitionen und Typdefinitonen ***********/
/******************************************************************/

#define ZEROOBJECT 32767
#define ABSTRACT = 0

DefClass(TCollection)
DefClass(TSortedCollection)
DefClass(TPersistent)
DefClass(TIRefStream)
DefClass(TORefStream)
DefClass(TIFRefStream)
DefClass(TOFRefStream)
DefClass(TAssociation)
DefClass(ostream)
DefClass(istream)
DefClass(ofstream)
DefClass(ifstream)

typedef void*		Pvoid;
typedef char*		Pchar;

typedef PTPersistent(*PTBuild)(PTIRefStream);
/*
 PTBuild ist der Name des Typs. Dieser Typ ist ein Pointer,
 der auf eine Funktion oder Methode zeigt, die einen Parameter
 vom Typ TIRefStream* erhaelt und einen Wert vom Typ PTPersistent
 zurueckliefert.
*/


/******************************************************************/
/********************** Error Codes *******************************/
/******************************************************************/

#define STREAM_OVERFLOW 1
/* Die Kapazitaet des Streams zur Verwaltung weiterer Instanzreferenzen
   ist erschoepft. */

#define WRONG_INTSIZE 2
/* Diese Bibliothek unterstuetzt nur 2- und 4-Byte Integer. Auf Maschinen
   mit anderen Werten entsteht dieser Fehler. */

#define LOAD_ERROR 3
/* Eine Instanz ist im Input-Stream als bereits geladen gekennzeichnet, es
   existiert aber kein Verweis auf diese Instanz (nur ein NULL-Pointer).
   Dieser Fehler kann eigentlich nur durch Implementationsfehler bei der
   Erweiterung dieser Bibliothek auftreten. */

#define CLASSMANAGER_ERR 4
/* Die Klasse eines zu ladenen Objekts ist dem ClassManager nicht bekannt.
   Eine wahrscheinliche Fehlerursache ist das Lesen eines falschen
   Klassennamens aus dem Stream oder, dass der Aufruf des Makros
   RegClass fuer diese Klasse unterblieben ist. */

#define INT_OVERFLOW 5
/* Dieser Fehler tritt auf, wenn ein 4-Byte Integerwert auf einer Maschine
   gelesen wird, die nur 2-Byte Integerwerte unterstuetzt. */

#define WRONG_COLL_INDEX 6
/* Beim Zugriff auf eine Collection wurde ein Indexwert ausserhalb des
   gueltigen Bereichs benutzt. */

#define COLL_OVERFLOW 7
/* Die maximale Anzahl an Elementen, die in eine Collection haben kann
   (normalerweise MaxCollectionSize = 16380), wurde ueberschritten. */

#define TASSOCIATION_ERR 8
/* Es wurde versucht, eine Instanz von TAssociation abzuspeichern;
   Instanzen dieser Klasse sind aber nicht persistenzfaehig. */


/******************************************************************/
/***** Abstrakte Basisklasse fuer alle Klassen mit ****************/
/***** persistenzfaehigen Instanzen                ****************/
/******************************************************************/
class TPersistent
{
  public:
      TPersistent(){};
      TPersistent(PTIRefStream is);
      virtual ~TPersistent(){};  

      // Abstrakte Store() und Classname() Methoden
      virtual void  Store(PTORefStream os) ABSTRACT;
      virtual Pchar Classname() ABSTRACT;
};


/******************************************************************/
/********* Die Streamklassen TORefStream und TIRefStream **********/
/******************************************************************/

class TRefStreamBase
{
  protected:
      PTCollection	KnownInstances;
      /* TCollection zur Verwaltung bereits gespeicherter Instanzen
	 in TORefStream bzw. zur Verwaltung bereits geladener Instanzen
	 in TIRefStream. */
};

class TORefStream : public TRefStreamBase
{
  public:
      TORefStream(Postream OS);
      virtual ~TORefStream();

      virtual void WriteInstanceConnection(PTPersistent objectpointer);
      /* Speichert Verweis auf ein Objekt und loest ggfs. das
	 Speichern des Objekts aus. */

      // Durchreicheoperationen zu den Standard-Streams:
      virtual void WriteInt(int value);
      virtual void WriteUnsigned(unsigned int value);
      virtual void WriteBytes(Pvoid source, int size);
      virtual void WriteDynamicString(Pchar source);
      virtual void WriteStaticString(Pchar source);

      friend TORefStream& operator<< (TORefStream& OS, char& source);
      friend TORefStream& operator<< (TORefStream& OS, unsigned char& source);
      friend TORefStream& operator<< (TORefStream& OS, short& source);
      friend TORefStream& operator<< (TORefStream& OS, unsigned short& source);
      friend TORefStream& operator<< (TORefStream& OS, int& source);
      friend TORefStream& operator<< (TORefStream& OS, unsigned int& source);
      friend TORefStream& operator<< (TORefStream& OS, long& source);
      friend TORefStream& operator<< (TORefStream& OS, unsigned long& source);
      friend TORefStream& operator<< (TORefStream& OS, float& source);
      friend TORefStream& operator<< (TORefStream& OS, double& source);

      friend TORefStream& operator<< (TORefStream& OS, PTPersistent source);

  protected:
      Postream os;
};
                       

class TIRefStream : public TRefStreamBase
{
  public:
      TIRefStream(Pistream IS);
      virtual ~TIRefStream();

      virtual PTPersistent ReadInstanceConnection();
      /* Gibt einen Verweis auf ein Objekt zurueck und laedt dieses
	 gegebenenfalls. Im Falle des Nachladens muss ReadInstanceConnection()
	 zuerst den Klassenname lesen, um ueber den ClassManager die
	 richtige Build-Methode aufrufen zu koennen. */

      virtual void EstablishInstance(PTPersistent objectpointer);
      /* Traegt Zeiger auf erstmals geladene Objekte in KnownInstances ein. */

      // Durchreicheoperationen zu den Standard-Streams:
      virtual void ReadBytes(Pvoid dest, int size);
      virtual void ReadInt(int& value);
      virtual void ReadUnsigned(unsigned int& value);
      virtual void ReadDynamicString(Pchar& dest);
      virtual void ReadStaticString(Pchar dest);

      friend TIRefStream& operator>> (TIRefStream& IS, char& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, unsigned char& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, short& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, unsigned short& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, int& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, unsigned int& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, long& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, unsigned long& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, float& dest);
      friend TIRefStream& operator>> (TIRefStream& IS, double& dest);

      friend TIRefStream& operator>> (TIRefStream& IS, PTPersistent& dest);

  protected:
      Pistream is;
};

/******************************************************************/
/********* Die Streamklassen TOFRefStream und TIFRefStream ********/
/******************************************************************/

class TOFRefStream : public TORefStream
{
  public:
      int isOpen;

      TOFRefStream(Pchar filename);
      virtual ~TOFRefStream();
      virtual void Close();
};


class TIFRefStream : public TIRefStream
{
  public:
      int isOpen;

      TIFRefStream(Pchar filename);
      virtual ~TIFRefStream();
      virtual void Close();
};

/******************************************************************/
/************ Definitionen fuer TCollection ***********************/
/******************************************************************/


typedef PTPersistent  ItemList[MaxCollectionSize];
typedef ItemList*     PItemList;

enum TOwnership {NotOwner,Owner};

typedef int  (*PTIterateFunc) (PTPersistent);
/* Typdefinition fuer die Parameter Funktionen von FirstThat(),
   LastThat() und ForEach() */

typedef void* TKey;
/* Schluessel-Typ bei TSortedCollection (wird z.B. in TStrKeyCollection
   genauer spezifiziert {Type-Cast notwendig!}) */


/******************************************************************/
/************ Klassen-Definition TCollection **********************/
/******************************************************************/

class TCollection : public TPersistent
{
  protected:
      int	  count;
      int	  delta;
      int	  limit;
      TOwnership  ownership;
      PItemList	  items;

      virtual void FreeItem(PTPersistent Item);

      DefNameBuild(TCollection)

  public:
      TCollection(int ALimit,int ADelta,TOwnership AnOwnership=NotOwner);
      TCollection(PTIRefStream is);
      virtual ~TCollection();

      virtual void Store(PTORefStream os);
      virtual int  Count(){return count;};
      virtual int  Delta(){return delta;};
      virtual int  Limit(){return limit;};
      virtual void SetOwnership(TOwnership mode){ownership=mode;};

      virtual TOwnership GetOwnership(){return ownership;};
      virtual PItemList  Items(){return items;};

      virtual void AtDelete(int Index);
      virtual void AtFree(int Index);
      virtual void AtInsert(int Index, PTPersistent Item);
      virtual void AtPut(int Index, PTPersistent Item);

      virtual PTPersistent At(int Index);

      virtual void DeleteAll();
      virtual void Delete(PTPersistent Item);
      virtual void Error(int Code);

      virtual PTPersistent FirstThat(PTIterateFunc Test);
      virtual void	   ForEach(PTIterateFunc Run);
      virtual PTPersistent LastThat(PTIterateFunc Test);

      virtual void FreeAll();
      virtual void Free(PTPersistent Item);
      virtual int  IndexOf(PTPersistent Item);
      virtual void Insert(PTPersistent Item);
      virtual void Pack();
      virtual void SetLimit(int ALimit);
};

 
/******************************************************************/
/********** Klassen-Definition TSortedCollection ******************/
/******************************************************************/

class TSortedCollection : public TCollection
{
  public:
      int	Duplicates;

      TSortedCollection(int ALimit,int ADelta,TOwnership AnOwnership=NotOwner);
      TSortedCollection(PTIRefStream is);

      virtual void Store(PTORefStream os);
      virtual int  Compare(TKey key1, TKey key2) ABSTRACT;
      virtual int  IndexOf(PTPersistent Item);
      virtual void Insert(PTPersistent Item);
      virtual TKey KeyOf(PTPersistent Item) ABSTRACT;
      virtual int  Search(TKey key, int& Index);
};


/******************************************************************/
/************ Definition von TAssociation *************************/
/******************************************************************/

class TAssociation : public TPersistent
{
  public:
      Pchar   IndexStr;
      Pvoid   AssociatedWith;

      TAssociation(Pchar indexstr, Pvoid valueptr);

      virtual void Store(PTORefStream OS)
      {
	cerr << "Error: TAssociation is not persistent!";
	exit(TASSOCIATION_ERR);
      };
      virtual Pchar Classname(){return "TAssociation";}
};
/* Element fuer TStrKeyCollection.
   Wird anhand dem String IndexStr in eine TStrKeyCollection
   einsortiert. Der Pointer AssociatedWith zeigt auf etwas, das mit dem
   String in Verbindung steht (z.B. weitere Datenfelder, ein Objekt oder
   eine Funktion)
*/

/******************************************************************/
/************ Definition von TStrKeyCollection ********************/
/******************************************************************/

class TStrKeyCollection : public TSortedCollection
{
      DefNameBuild(TStrKeyCollection)
  public:
      TStrKeyCollection(int ALimit,int ADelta,TOwnership AnOwnership=NotOwner);
      TStrKeyCollection(PTIRefStream is);

      virtual int  Compare(TKey key1, TKey key2);
      virtual TKey KeyOf(PTPersistent Item);
};


#endif
