#include "streampp.h"
    
/******************************************************************/
/******** Anlegen des ClassManagers *******************************/
/******************************************************************/

// Der ClassManager ist eine Datenstruktur mit Key-Zugriff, die bei
// Angabe des Klassennamens einen Zeiger auf die passende Build-Methode
// zurueckgibt.
// Eintraege werden in der TCollection ClassManager durch den 
// Makroaufruf RegClass(name) erstellt.
TStrKeyCollection	ClassManager(50,10);


/******************************************************************/
/******** Implementation der Klasse TPersistent *******************/
/******************************************************************/


TPersistent::TPersistent(PTIRefStream is)
{
  is->EstablishInstance(this);
}


/******************************************************************/
/***************** Methoden von TORefStream ***********************/
/******************************************************************/

TORefStream::TORefStream(Postream OS)
{
  KnownInstances = new TCollection(250,50);
  os = OS;
}


TORefStream::~TORefStream()
{
  KnownInstances->DeleteAll();
  delete KnownInstances;
}


void TORefStream::WriteInstanceConnection(PTPersistent objectpointer)
{
  int found=0;
  int size=KnownInstances->Count();
  int index;

  if (objectpointer==NULL)
  {
    // id fuer NULL Pointer wird abgespeichert und WriteInstanceConnection wird beendet.
    index = ZEROOBJECT;
    WriteInt(index);
    return;
  }

  for(index=0; (index<size)&&(!found);index++)
    if((KnownInstances->At(index)) == (PTPersistent) objectpointer) found=1;
  index--;

  if(found)
    WriteInt(index);
  else
  {
    KnownInstances->Insert((PTPersistent)objectpointer);
    // Ist die Kapazitaet von KnownInstances erschoepft?
    if (size == ZEROOBJECT)
    {
      cerr << "Stream overflow error!";
      exit(STREAM_OVERFLOW);
    }
    WriteInt(size); // id mit 0 beginnend
    WriteDynamicString(objectpointer->Classname());
    objectpointer->Store(this);
  }
}


void TORefStream::WriteBytes(Pvoid source, int size)
{
#ifdef ReverseByteSequencesForPC
  for (int i=size-1;i>=0;i--)
    os->write((((Pchar)source)+i),1);
#endif

#ifndef ReverseByteSequencesForPC     
  os->write((Pchar)source,size);
#endif
}

void TORefStream::WriteInt(int value)
{
  if (sizeof(int)==4)
  {
    WriteBytes(&value,4);
    return;
  }                 
  if (sizeof(int)==2)
  {
    long int dummy = value;                           
    WriteBytes(&dummy,4);
    return;
  }
  cerr << "This library does only support 2- or 4-byte integers!\n";
  exit(WRONG_INTSIZE);
}

void TORefStream::WriteUnsigned(unsigned int value)
{
  if (sizeof(unsigned int)==4)
  {
    WriteBytes(&value,4);
    return;
  }
  if (sizeof(unsigned int)==2)
  {
    unsigned long int dummy = value;
    WriteBytes(&dummy,4);
    return;
  }
  cerr << "This library does only support 2- or 4-byte integers!\n";
  exit(WRONG_INTSIZE);
}


void TORefStream::WriteDynamicString(Pchar source)
{
  int size = strlen(source)+1;
    // Abschliessende \0 wird mitgespeichert
  WriteInt(size);
  os->write((Pchar)source,size);
}

void TORefStream::WriteStaticString(Pchar source)
{
  WriteDynamicString(source);
}


TORefStream& operator<< (TORefStream& OS, char& source)
{
  OS.WriteBytes(&source,sizeof(char));
  return OS;
}

TORefStream& operator<< (TORefStream& OS, unsigned char& source)
{
  OS.WriteBytes(&source,sizeof(unsigned char));
  return OS;
}

TORefStream& operator<< (TORefStream& OS, short& source)
{
  OS.WriteBytes(&source,sizeof(short));
  return OS;
}

TORefStream& operator<< (TORefStream& OS, unsigned short& source)
{
  OS.WriteBytes(&source,sizeof(unsigned short));
  return OS;
}

TORefStream& operator<< (TORefStream& OS, int& source)
{
  OS.WriteInt(source);
  return OS;
}

TORefStream& operator<< (TORefStream& OS, unsigned int& source)
{
  OS.WriteUnsigned(source);
  return OS;
}

TORefStream& operator<< (TORefStream& OS, long& source)
{
  OS.WriteBytes(&source,sizeof(long));
  return OS;
}

TORefStream& operator<< (TORefStream& OS, unsigned long& source)
{
  OS.WriteBytes(&source,sizeof(unsigned long));
  return OS;
}


TORefStream& operator<< (TORefStream& OS, float& source)
{
  OS.WriteBytes(&source,sizeof(float));
  return OS;
}

TORefStream& operator<< (TORefStream& OS, double& source)
{
  OS.WriteBytes(&source,sizeof(double));
  return OS;
}


TORefStream& operator<< (TORefStream& OS, PTPersistent source)
{
  OS.WriteInstanceConnection(source);
  return OS;
}


/******************************************************************/
/***************** Methoden von TIRefStream ***********************/
/******************************************************************/
TIRefStream::TIRefStream(Pistream IS)
{
  KnownInstances = new TCollection(250,50);
  is = IS;
}


TIRefStream::~TIRefStream()
{ 
  KnownInstances->DeleteAll();
  delete KnownInstances;
}


void TIRefStream::EstablishInstance(PTPersistent objectpointer)
{
  KnownInstances->Insert(objectpointer);
}


PTPersistent TIRefStream::ReadInstanceConnection()
{
  int  		id;
  Pchar 	classname;
  int		ClassManagerIndex;
  PTPersistent	PersObject;
  PTAssociation AssocObject;
  PTBuild	BuildMethod;


  //Lesen der id
  ReadInt(id);

  //Abpruefen auf NULL-id
  if (id==ZEROOBJECT) return NULL;

  //Abpruefen, ob Instanz schon geladen
  if (id < KnownInstances->Count())
  {
    if((KnownInstances->At(id)) == NULL)
    {
      cerr << "Fatal error when loading data!\n";
      exit(LOAD_ERROR);
    }
    return KnownInstances->At(id);
  }
  ReadDynamicString(classname);
  /* Aufruf der Build-Methode ueber den ClassManager und Eintragen eines
     Zeigers auf das neue Objekt in KnownInstances (an der Stelle,
     die der gelesenen id entspricht). */

  if(!(ClassManager.Search(classname,ClassManagerIndex)))
  {
    cerr << "Fatal error in ClassManager!\n";
    exit(CLASSMANAGER_ERR);
  }
  AssocObject = (PTAssociation)(ClassManager.At(ClassManagerIndex));
  BuildMethod = (PTBuild)(AssocObject->AssociatedWith);
  PersObject = BuildMethod(this);

  return PersObject;
  // oder return (*KnownInstances)[id];
}


void TIRefStream::ReadBytes(Pvoid dest, int size)
{
#ifdef ReverseByteSequencesForPC
  for (int i=size-1;i>=0;i--)
    is->read(((Pchar)dest)+i,1);
#endif

#ifndef ReverseByteSequencesForPC
  is->read((Pchar)dest,size);
#endif  
}

void TIRefStream::ReadInt(int &value)
{
  if (sizeof(int)==4)
  {
    ReadBytes(&value,4);
    return;
  }
  if (sizeof(int)==2)
  {
    long int testbytes;
    ReadBytes(&testbytes,4);

    if( (testbytes>32767) || (testbytes<-32768) )
    {
      cerr << "Integer overflow when loading data ('int' has only 2 bytes on this machine)\n";
      exit(INT_OVERFLOW);
    }
    value = testbytes;
    return;
  }
  cerr << "This library does only support 2- or 4-byte integers!\n";
  exit(WRONG_INTSIZE);
}


void TIRefStream::ReadUnsigned(unsigned int &value)
{
  if (sizeof(unsigned int)==4)
  {
    ReadBytes(&value,4);
    return;
  }
  if (sizeof(unsigned int)==2)
  {
    long int testbytes;
    ReadBytes(&testbytes,4);
    if(testbytes>65535)
    {
      cerr << "Integer overflow when loading data ('int' has only 2 bytes on this machine)\n";
      exit(INT_OVERFLOW);
    }
    value = testbytes;
    return;
  }
  cerr << "This library does only support 2- or 4-byte integers!\n";
  exit(WRONG_INTSIZE);
}



void TIRefStream::ReadDynamicString(Pchar &dest)
{
  int size;

  ReadInt(size);
  dest =(Pchar) malloc(size);
  is->read((Pchar)dest,size);
}

void TIRefStream::ReadStaticString(Pchar dest)
{
  int size;

  ReadInt(size);
  is->read((Pchar)dest,size);
}


TIRefStream& operator>> (TIRefStream& IS, char& dest)
{
  IS.ReadBytes(&dest,sizeof(char));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, unsigned char& dest)
{
  IS.ReadBytes(&dest,sizeof(unsigned char));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, short& dest)
{
  IS.ReadBytes(&dest,sizeof(short));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, unsigned short& dest)
{
  IS.ReadBytes(&dest,sizeof(unsigned short));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, int& dest)
{
  IS.ReadInt(dest);
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, unsigned int& dest)
{
  IS.ReadUnsigned(dest);
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, long& dest)
{
  IS.ReadBytes(&dest,sizeof(long));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, unsigned long& dest)
{
  IS.ReadBytes(&dest,sizeof(unsigned long));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, float& dest)
{
  IS.ReadBytes(&dest,sizeof(float));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, double& dest)
{
  IS.ReadBytes(&dest,sizeof(double));
  return IS;
}

TIRefStream& operator>> (TIRefStream& IS, PTPersistent& dest)
{
  dest = IS.ReadInstanceConnection();
  return IS;
}

/******************************************************************/
/********* Die Methoden von TOFRefStream und TIFRefStream *********/
/******************************************************************/

TOFRefStream::TOFRefStream(Pchar filename):TORefStream(NULL)
{
  os = new ofstream();
  ((Pofstream)os)->open(filename,FileOutOpenMode,filebuf::openprot);
  isOpen=1;
}

TOFRefStream::~TOFRefStream()
{
  Close();
}

void TOFRefStream::Close()
{
  if(isOpen)
  {
    ((Pofstream)os)->close();
    delete os;
  }
  isOpen=0;
}

TIFRefStream::TIFRefStream(Pchar filename):TIRefStream(NULL)
{
  is= new ifstream();
  ((Pifstream)is)->open(filename,FileInOpenMode,filebuf::openprot);
  isOpen=1;
}

TIFRefStream::~TIFRefStream()
{
  Close();
}

void TIFRefStream::Close()
{
  if(isOpen)
  {
    ((Pifstream)is)->close();
    delete is;
  }
  isOpen=0;
}


/******************************************************************/
/******** Implementierung der Methoden von TCollection ************/
/******************************************************************/

TCollection::TCollection(int ALimit,int ADelta,TOwnership AnOwnership):TPersistent()
{
  limit=ALimit;
  delta=ADelta;
  ownership=AnOwnership;
  count=0;
  items =  (PItemList) malloc(limit * sizeof(PTPersistent));
}

// Build-Konstruktor fuer die RefStream-Faehigkeit der TCollection
TCollection::TCollection(PTIRefStream is):TPersistent(is)
{  
  int temp;
  is->ReadInt(count);
  is->ReadInt(delta);
  is->ReadInt(limit);
  is->ReadInt(temp);
  ownership = (TOwnership) temp;
  
  items = (PItemList) malloc(limit * sizeof(PTPersistent));
  for (int i=0;i<count;i++) (*items)[i] =(PTPersistent) is->ReadInstanceConnection();
}


TCollection::~TCollection()
{
  if (ownership==Owner) FreeAll();
  free((char*)items);
  /* "char*" deshalb, weil auf einigen Maschinen free und malloc nur mit
     diesem Parameter bzw. Rueckgabewert definiert sind. */
}

// Store-Methode fuer persistenzfaehige Instanzen von TCollection
void TCollection::Store(PTORefStream os)
{
  os->WriteInt(count);
  os->WriteInt(delta);
  os->WriteInt(limit);
  os->WriteInt(ownership);

  for (int i=0;i<count;i++) os->WriteInstanceConnection((PTPersistent)(*items)[i]);
}


// Entfernt Element aus der Collection
void TCollection::AtDelete(int Index)
{
  if ((Index>=count)||(Index<0)) Error(1);
  (*items)[Index]=NULL;
  Pack();
}

// Entfernt Element aus der Collection und loescht es
void TCollection::AtFree(int Index)
{
  FreeItem(At(Index));
  AtDelete(Index);
}

// Fuegt Element in die Collection ein
void TCollection::AtInsert(int Index, PTPersistent Item)
{
  if ((Index>count)||(Index<0)) Error(1);
  count++;
  if (count==limit)
  {
    if ((MaxCollectionSize-delta)<limit)
      Error(2);
    else
      SetLimit(limit+delta);
  }
  for(int i=count-1;i>Index;i--) (*items)[i] = At(i-1);
  (*items)[Index] = Item;
}


// Ueberschreibt einen Eintrag in der Collection
void TCollection::AtPut(int Index, PTPersistent Item)
{
  if ((Index>=count)||(Index<0)) Error(1);
  (*items)[Index] = Item;
}


// Gibt einen Zeiger auf das entsprechende Objekt in der Collection zurueck
PTPersistent TCollection::At(int Index)
{
  if ((Index>=count)||(Index<0)) Error(1);
  return (*items)[Index];
}


// Entfernt alle Eintraege aus der Collection
void TCollection::DeleteAll()
{
  for(int i=0;i<count;i++) (*items)[i] = NULL;
  count=0;
}


// Entfernt den Eintrag des entsprechenden Objekts aus der Collection
void TCollection::Delete(PTPersistent Item)
{
  AtDelete(IndexOf(Item));
}


// Methode zur Fehlerbehandlung
void TCollection::Error(int Code)
{
  switch (Code)
  {
    case 1: cerr<< "\nCollection-index out of range !\n"; exit(WRONG_COLL_INDEX);
    case 2: cerr<< "\nCollection overflow !\n"; exit(COLL_OVERFLOW);
  }
}

// Suchen in Collection (von vorne)
PTPersistent TCollection::FirstThat(PTIterateFunc Test)
{
  PTPersistent Ptemp=NULL;
  for(int i=0;(i<count)&&(Ptemp==NULL);i++)
    if (Test(At(i))) Ptemp = At(i);
  return Ptemp;
}

// Iteriert ueber Collection
void TCollection::ForEach(PTIterateFunc Run)
{
  for(int i=0;(i<count);i++) Run(At(i));
}


// Suchen in Collection (von hinten)
PTPersistent TCollection::LastThat(PTIterateFunc Test)
{
  PTPersistent Ptemp=NULL;
  for(int i=count-1;(i>=0)&&(Ptemp==NULL);i--)
    if (Test(At(i))) Ptemp = At(i);
  return Ptemp;
}


// Entfernt alle Eintraege aus der Collection und
// loescht die dazugehoerenden Objekte
void TCollection::FreeAll()
{
  for(int i=0;i<count;i++)
  {
    FreeItem(At(i));
    (*items)[i]=NULL;
  }
  count = 0;
}


// Loescht ein Objekt
void TCollection::FreeItem(PTPersistent Item)
{
  if (Item != NULL) delete Item;
}


// Loescht ein Objekt und entfernt den
// entsprechenden Eintrag aus der Collection
void TCollection::Free(PTPersistent Item)
{
  Delete(Item);
  FreeItem(Item);
}
 	

// Liefert die Indexnummer des Objekts in der Collection
// -1 falls nicht in Collection
int TCollection::IndexOf(PTPersistent Item)
{
  for(int i=0;i<count;i++)
    if (At(i) == Item) return i;
  return -1;
}


// Fuegt ein Element in die Collection ein
void TCollection::Insert(PTPersistent Item)
{
  AtInsert(count,Item);
}


// Entfernt NULL Zeiger aus der Collection
void TCollection::Pack()
{
  for(int j=0;j<count;j++)
    if (At(j)==NULL)
    {
      for(int i=j;i<count-1;i++) (*items)[i] = At(i+1);
      // evtl. (*items)[count]=NULL;
      count--;
      j--;
    }
}


// Aendert die Array-Groesse dynamisch
void TCollection::SetLimit(int ALimit)
{
  PItemList Ptemp;

  if (ALimit>MaxCollectionSize)
    ALimit = MaxCollectionSize;

  Ptemp = (PItemList) malloc(ALimit * sizeof(PTPersistent));
  for(int i=0; i<count; i++)
    (*Ptemp)[i] = At(i);

  free((char*)items);
  items=Ptemp;
  limit = ALimit;
}
 


/******************************************************************/
/***** Implementierung der Methoden von TSortedCollection *********/
/******************************************************************/

TSortedCollection::TSortedCollection(int ALimit,int ADelta,TOwnership AnOwnership)
	:TCollection(ALimit,ADelta,AnOwnership)
{
  Duplicates = 0;
}


TSortedCollection::TSortedCollection(PTIRefStream is):TCollection(is)
{
  is->ReadInt(Duplicates);
}


void TSortedCollection::Store(PTORefStream os)
{
  TCollection::Store(os);
  os->WriteInt(Duplicates);
}


int  TSortedCollection::IndexOf(PTPersistent Item)
{
  int index;

  if (!Search(KeyOf(Item),index))
    return -1;
  else
    return index;
}


void TSortedCollection::Insert(PTPersistent Item)
{
  int index;

  if(Search(KeyOf(Item),index) && !Duplicates)
    AtPut(index,Item);
  else
    AtInsert(index,Item);
}


// Binaeres Suchen
int TSortedCollection::Search(TKey Key, int& Index)
{

  int upperbound=count-1,lowerbound=0,i=1;
  Index=0;

  if (!count) return 0;

  while(( upperbound >= lowerbound )&&(i))
  {
    i=Compare(Key,KeyOf((*items)[lowerbound+(upperbound-lowerbound)/2]));
    if (i>0) lowerbound = lowerbound + (upperbound-lowerbound)/2+1;
    if (i<0) upperbound = lowerbound + (upperbound-lowerbound)/2-1;
  }

  if (i)
  {
    Index = lowerbound;
    return 0;
  }
  else
  {
    Index = lowerbound+(upperbound-lowerbound)/2;
    return 1;
  }
}


/******************************************************************/
/******** Implementation der Klasse TStrKeyCollection *************/
/******************************************************************/

TStrKeyCollection::TStrKeyCollection(int ALimit,int ADelta,TOwnership AnOwnership)
	:TSortedCollection(ALimit,ADelta,AnOwnership)
{
}


TStrKeyCollection::TStrKeyCollection(PTIRefStream is)
	:TSortedCollection(is)
{
}


int TStrKeyCollection::Compare(TKey key1, TKey key2)
{
  return strcmp((Pchar)key1,(Pchar)key2);
}


TKey TStrKeyCollection::KeyOf(PTPersistent Item)
{
  return (TKey)(((PTAssociation)Item)->IndexStr);
}



/******************************************************************/
/******** Implementation der Klasse TAssociation ******************/
/******************************************************************/

TAssociation::TAssociation(Pchar indexstr, Pvoid associatedwith)
{
  AssociatedWith = associatedwith;
  IndexStr = (char*)malloc(strlen(indexstr)+1);
  strcpy(IndexStr,indexstr); 
}
